import { ifNull, isFunction, isIterable, isPlainObject, Nullable, orElse } from '../utils'
import { ClassSerializer, IterableSerializer, ObjectSerializer, Serializable, Serializer } from './interface'

const unhandledType = (serializer: any, value: any) => new TypeError(`'${serializer.constructor.name}' has no method for value type '${typeof value}'`)

function serializeEntries<T, K extends Serializable, V extends Serializable, E extends Iterable<[K, V]>>(serializer: ObjectSerializer<T>, value: E) {
  let state

  for (const [key, val] of value) {
    state = serializer.serializeKey(key)
    state = serializer.serializeValue(val)
  }

  return serializer.end()
}

function serializeObject<T, K extends PropertyKey, V extends Serializable, R extends Record<K, V>>(serializer: ObjectSerializer<T>, value: R) {
  return serializeEntries(serializer, Object.entries(value) as Iterable<[K, V]>)
}

function getClassName(value: any): Nullable<string> {
  return value?.constructor.name
}

function serializeClass<T, K extends PropertyKey, V extends Serializable, R extends Record<K, V>>(serializer: ClassSerializer<T>, value: R) {
  for (const prop in value) {
    serializer.serializeField(prop, value[prop])
  }

  return serializer.end()
}

function serializeIter<T, V extends Iterable<any>>(serializer: IterableSerializer<T>, value: V) {
  let state

  for (const val of value) {
    state = serializer.serializeElement(val)
  }

  return serializer.end()
}

// dispatches in the order of serialize<type> -> serializeAny -> throw TypeError
export function serializeWith<T>(serializer: Serializer<T>, value: Serializable): Nullable<T> {
  // prepare fallback methods
  const serializeAny = orElse(
    serializer,
    serializer.serializeAny,
    (value: Serializable) => unhandledType(serializer, value)
  )

  const serialize = ifNull(serializer, serializeAny, value)

  switch (typeof value) {
    case 'string': return serialize(serializer.serializeString)
    case 'number': return serialize(serializer.serializeNumber)
    case 'bigint': return serialize(serializer.serializeBigInt)
    case 'boolean': return serialize(serializer.serializeBoolean)
    case 'symbol': return serialize(serializer.serializeSymbol)
    case 'undefined': return serialize(serializer.serializeNull)
    case 'function': return serialize(serializer.serializeFunction)

    case 'object':
      if (value instanceof Map && isFunction(serializer.serializeMap)) {
        return serializeEntries(serializer.serializeMap(), value)
      } else if (isIterable(value) && isFunction(serializer.serializeIterable)) {
        return serializeIter(serializer.serializeIterable(), value)
      } else if (isFunction(serializer.serializeClass) && !isPlainObject(value)) {
        const name = getClassName(value)
        return serializeClass(serializer.serializeClass(name!), value as any)
      } else if (isFunction(serializer.serializeObject)) {
        return serializeObject(serializer.serializeObject!(), value as Record<PropertyKey, any>)
      } // deliberate fallthrough when the above fail

    default: return serializeAny(value)
  }
}

