export type Nullable<T> = T | undefined

export type Primitive = string | number | boolean | symbol | bigint | null | undefined

export interface ToString {
  toString(): string
}

export function staticImplements<T>() {
  return <U extends T>(constructor: U) => { constructor }
}

export function isPlainObject(value: any): boolean {
  return Object.getPrototypeOf(value) === Object.prototype
}

export function isFunction(value: any): value is Function {
  return value != null && typeof value === 'function'
}

export function isIterable(value: any): value is Iterable<any> {
  return isFunction(value[Symbol.iterator])
}

export type Constructor<T = any> = new (...args: any[]) => T

export const CaseConvention = Object.freeze({
  Lowercase: 0,
  Uppercase: 1,
  PascalCase: 2,
  CamelCase: 3,
  SnakeCase: 4,
  ScreamingSnakeCase: 5,
  KebabCase: 6,
  ScreamingKebabCase: 7
} as const)

export type CaseConvention = typeof CaseConvention[keyof typeof CaseConvention]

export function orElse(thisArg: any, a: Nullable<Function>, b: Function) {
  return function(...args: any) {
    const fn = a != null ? a : b
    return fn.apply(thisArg, args)
  }
}

export function ifNull(thisArg: any, b: Function, ...args: any) {
  return function(a: Nullable<Function>) {
    return orElse(thisArg, a, b).call(thisArg, args)
  }
}

function applyMixins(derivedCtor: any, constructors: any[]) {
  constructors.forEach((baseCtor) => {
    Object.getOwnPropertyNames(baseCtor.prototype).forEach((name) => {
      Object.defineProperty(
        derivedCtor.prototype,
        name,
        Object.getOwnPropertyDescriptor(baseCtor.prototype, name) ||
        Object.create(null)
      )
    })
  })
}

export function mixin<U = any>(impl: Function) {
  return function <TBase extends Constructor<U>>(constructor: TBase) {
    applyMixins(constructor, [impl])
    return constructor
  }
}

